@@ -16,6 +16,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <jni.h>
+
 #ifdef linux
 #include <dlfcn.h>
 #endif
@@ -42,6 +43,7 @@ static void t_jccenv_dealloc(t_jccenv *self);
 static PyObject *t_jccenv_attachCurrentThread(PyObject *self, PyObject *args);
 static PyObject *t_jccenv_detachCurrentThread(PyObject *self);
 static PyObject *t_jccenv_isCurrentThreadAttached(PyObject *self);
+static PyObject *t_jccenv_isShared(PyObject *self);
 static PyObject *t_jccenv_strhash(PyObject *self, PyObject *arg);
 static PyObject *t_jccenv__dumpRefs(PyObject *self,
                                     PyObject *args, PyObject *kwds);
@@ -69,6 +71,8 @@ static PyMethodDef t_jccenv_methods[] = {
       METH_NOARGS, NULL },
     { "isCurrentThreadAttached", (PyCFunction) t_jccenv_isCurrentThreadAttached,
       METH_NOARGS, NULL },
+    { "isShared", (PyCFunction) t_jccenv_isShared,
+      METH_NOARGS, NULL },
     { "strhash", (PyCFunction) t_jccenv_strhash,
       METH_O, NULL },
     { "_dumpRefs", (PyCFunction) t_jccenv__dumpRefs,
@@ -110,7 +114,6 @@ PyTypeObject PY_TYPE(JCCEnv) = {
     t_jccenv_methods,                    /* tp_methods */
     t_jccenv_members,                    /* tp_members */
     t_jccenv_properties,                 /* tp_getset */
-    0,                                   /* tp_getset */
     0,                                   /* tp_base */
     0,                                   /* tp_dict */
     0,                                   /* tp_descr_get */
@@ -154,21 +157,11 @@ static PyObject *t_jccenv_attachCurrentThread(PyObject *self, PyObject *args)
 {
     char *name = NULL;
     int asDaemon = 0, result;
-    JNIEnv *jenv = NULL;
 
     if (!PyArg_ParseTuple(args, "|si", &name, &asDaemon))
         return NULL;
 
-    JavaVMAttachArgs attach = {
-        JNI_VERSION_1_4, name, NULL
-    };
-
-    if (asDaemon)
-        result = env->vm->AttachCurrentThreadAsDaemon((void **) &jenv, &attach);
-    else
-        result = env->vm->AttachCurrentThread((void **) &jenv, &attach);
-
-    env->set_vm_env(jenv);
+    result = env->attachCurrentThread(name, asDaemon);
         
     return PyInt_FromLong(result);
 }
@@ -190,6 +183,15 @@ static PyObject *t_jccenv_isCurrentThreadAttached(PyObject *self)
     Py_RETURN_FALSE;
 }
 
+static PyObject *t_jccenv_isShared(PyObject *self)
+{
+#ifdef _jcc_lib
+    Py_RETURN_TRUE;
+#else
+    Py_RETURN_FALSE;
+#endif
+}
+
 static PyObject *t_jccenv_strhash(PyObject *self, PyObject *arg)
 {
     int hash = PyObject_Hash(arg);
@@ -330,7 +332,7 @@ static void registerNatives(JNIEnv *vm_env);
 _DLL_EXPORT PyObject *initJCC(PyObject *module)
 {
     static int _once_only = 1;
-#ifdef _MSC_VER
+#if defined(_MSC_VER) || defined(__WIN32)
 #define verstring(n) #n
     PyObject *ver = PyString_FromString(verstring(JCC_VER));
 #else
@@ -361,9 +363,9 @@ _DLL_EXPORT PyObject *initVM(PyObject *self, PyObject *args, PyObject *kwds)
     };
     char *classpath = NULL;
     char *initialheap = NULL, *maxheap = NULL, *maxstack = NULL;
-    char *vmargs = NULL;
+    PyObject *vmargs = NULL;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|zzzzz", kwnames,
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|zzzzO", kwnames,
                                      &classpath,
                                      &initialheap, &maxheap, &maxstack,
                                      &vmargs))
@@ -441,14 +443,18 @@ _DLL_EXPORT PyObject *initVM(PyObject *self, PyObject *args, PyObject *kwds)
         if (maxstack)
             add_option("-Xss", maxstack, &vm_options[nOptions++]);
 
-        if (vmargs)
+        if (vmargs != NULL && PyString_Check(vmargs))
         {
-            char *buf = strdup(vmargs);
+#ifdef _MSC_VER
+            char *buf = _strdup(PyString_AS_STRING(vmargs));
+#else
+            char *buf = strdup(PyString_AS_STRING(vmargs));
+#endif
             char *sep = ",";
             char *option;
 
-            for (option = strtok(buf, sep); option; option = strtok(NULL, sep))
-            {
+            for (option = strtok(buf, sep); option != NULL;
+                 option = strtok(NULL, sep)) {
                 if (nOptions < sizeof(vm_options) / sizeof(JavaVMOption))
                     add_option("", option, &vm_options[nOptions++]);
                 else
@@ -456,13 +462,59 @@ _DLL_EXPORT PyObject *initVM(PyObject *self, PyObject *args, PyObject *kwds)
                     free(buf);
                     for (unsigned int i = 0; i < nOptions; i++)
                         delete vm_options[i].optionString;
-                    PyErr_Format(PyExc_ValueError, "Too many options (> %d)",
-                                 nOptions);
+                    PyErr_Format(PyExc_ValueError,
+                                 "Too many options (> %d)", nOptions);
                     return NULL;
                 }
             }
             free(buf);
         }
+        else if (vmargs != NULL && PySequence_Check(vmargs))
+        {
+            PyObject *fast =
+                PySequence_Fast(vmargs, "error converting vmargs to a tuple");
+
+            if (fast == NULL)
+                return NULL;
+
+            for (int i = 0; i < PySequence_Fast_GET_SIZE(fast); ++i) {
+                PyObject *arg = PySequence_Fast_GET_ITEM(fast, i);
+
+                if (PyString_Check(arg))
+                {
+                    char *option = PyString_AS_STRING(arg);
+
+                    if (nOptions < sizeof(vm_options) / sizeof(JavaVMOption))
+                        add_option("", option, &vm_options[nOptions++]);
+                    else
+                    {
+                        for (unsigned int j = 0; j < nOptions; j++)
+                            delete vm_options[j].optionString;
+                        PyErr_Format(PyExc_ValueError,
+                                     "Too many options (> %d)", nOptions);
+                        Py_DECREF(fast);
+                        return NULL;
+                    }
+                }
+                else
+                {
+                    for (unsigned int j = 0; j < nOptions; j++)
+                        delete vm_options[j].optionString;
+                    PyErr_Format(PyExc_TypeError,
+                                 "vmargs arg %d is not a string", i);
+                    Py_DECREF(fast);
+                    return NULL;
+                }
+            }
+
+            Py_DECREF(fast);
+        }
+        else if (vmargs != NULL)
+        {
+            PyErr_SetString(PyExc_TypeError,
+                            "vmargs is not a string or sequence");
+            return NULL;
+        }
 
         //vm_options[nOptions++].optionString = "-verbose:gc";
         //vm_options[nOptions++].optionString = "-Xcheck:jni";
@@ -497,6 +549,57 @@ _DLL_EXPORT PyObject *initVM(PyObject *self, PyObject *args, PyObject *kwds)
     }
 }
 
+/* returns borrowed reference */
+_DLL_EXPORT PyObject *getJavaModule(PyObject *module,
+                                    const char *parent, const char *name) {
+    PyObject *modules = PyImport_GetModuleDict();
+    PyObject *parent_module, *full_name;
+
+    if (parent[0] == '\0')
+    {
+        parent_module = NULL;
+        full_name = PyString_FromString(name);
+    }
+    else if ((parent_module = PyDict_GetItemString(modules, parent)) == NULL)
+    {
+        PyErr_Format(PyExc_ValueError, "Parent module '%s' not found", parent);
+        return NULL;
+    }
+    else
+        full_name = PyString_FromFormat("%s.%s", parent, name);
+
+    PyObject *child_module = PyDict_GetItem(modules, full_name);
+
+    if (child_module == NULL)
+    {
+        child_module = PyModule_New(PyString_AS_STRING(full_name));
+        if (child_module != NULL)
+        {
+            if (parent_module != NULL)
+                PyDict_SetItemString(PyModule_GetDict(parent_module),
+                                     name, child_module);
+            PyDict_SetItem(modules, full_name, child_module);
+            Py_DECREF(child_module);  /* borrow reference */
+        }
+    }
+    Py_DECREF(full_name);
+
+    /* During __install__ pass, __file__ is not yet set on module.
+     * During __initialize__ pass, __file__ is passed down to child_module.
+     */
+    if (child_module != NULL)
+    {
+        PyObject *__file__ = PyString_FromString("__file__");
+        PyObject *file = PyDict_GetItem(PyModule_GetDict(module), __file__);
+
+        if (file != NULL)
+            PyDict_SetItem(PyModule_GetDict(child_module), __file__, file);
+        Py_DECREF(__file__);
+    }
+
+    return child_module;
+}
+
 #ifdef _jcc_lib
 
 static void raise_error(JNIEnv *vm_env, const char *message)
@@ -659,68 +762,160 @@ extern "C" {
     }
 };
 
-static void JNICALL _PythonException_getErrorInfo(JNIEnv *vm_env, jobject self)
+static void JNICALL _PythonException_pythonDecRef(JNIEnv *vm_env, jobject self)
+{
+    jclass jcls = vm_env->GetObjectClass(self);
+    jfieldID fid = vm_env->GetFieldID(jcls, "py_error_state", "J");
+    PyObject *state = (PyObject *) vm_env->GetLongField(self, fid);
+    
+    if (state != NULL)
+    {
+        PythonGIL gil(vm_env);
+
+        Py_DECREF(state);
+        vm_env->SetLongField(self, fid, (jlong) 0);
+    }
+}
+
+static void JNICALL _PythonException_saveErrorState(JNIEnv *vm_env,
+                                                    jobject self)
 {
     PythonGIL gil(vm_env);
-
-    if (!PyErr_Occurred())
-        return;
-
-    PyObject *type, *value, *tb, *errorName;
-    jclass jcls = vm_env->GetObjectClass(self);
+    PyObject *type, *value, *tb;
 
     PyErr_Fetch(&type, &value, &tb);
 
-    errorName = PyObject_GetAttrString(type, "__name__");
+    if (type != NULL)
+    {
+        PyObject *state = PyTuple_New(3);
+
+        PyErr_NormalizeException(&type, &value, &tb);
+        PyTuple_SET_ITEM(state, 0, type);
+        if (value == NULL)
+        {
+            PyTuple_SET_ITEM(state, 1, Py_None);
+            Py_INCREF(Py_None);
+        }
+        else
+            PyTuple_SET_ITEM(state, 1, value);
+        if (tb == NULL)
+        {
+            PyTuple_SET_ITEM(state, 2, Py_None);
+            Py_INCREF(Py_None);
+        }
+        else
+            PyTuple_SET_ITEM(state, 2, tb);
+
+        jclass jcls = vm_env->GetObjectClass(self);
+        jfieldID fid = vm_env->GetFieldID(jcls, "py_error_state", "J");
+
+        vm_env->SetLongField(self, fid, (jlong) state);
+    }
+}
+
+static jstring JNICALL _PythonException_getErrorName(JNIEnv *vm_env,
+                                                     jobject self)
+{
+    jclass jcls = vm_env->GetObjectClass(self);
+    jfieldID fid = vm_env->GetFieldID(jcls, "py_error_state", "J");
+    PyObject *state = (PyObject *) vm_env->GetLongField(self, fid);
+    
+    if (state == NULL)
+        return NULL;
+
+    PythonGIL gil(vm_env);
+    PyObject *errorName =
+        PyObject_GetAttrString(PyTuple_GET_ITEM(state, 0), "__name__");
+
     if (errorName != NULL)
     {
-        jfieldID fid =
-            vm_env->GetFieldID(jcls, "errorName", "Ljava/lang/String;");
         jstring str = env->fromPyString(errorName);
-
-        vm_env->SetObjectField(self, fid, str);
-        vm_env->DeleteLocalRef(str);
         Py_DECREF(errorName);
+
+        return str;
     }
 
-    if (value != NULL)
+    return NULL;
+}
+
+static jstring JNICALL _PythonException_getErrorMessage(JNIEnv *vm_env,
+                                                        jobject self)
+{
+    jclass jcls = vm_env->GetObjectClass(self);
+    jfieldID fid = vm_env->GetFieldID(jcls, "py_error_state", "J");
+    PyObject *state = (PyObject *) vm_env->GetLongField(self, fid);
+    
+    if (state == NULL)
+        return NULL;
+
+    PythonGIL gil(vm_env);
+    PyObject *value = PyTuple_GET_ITEM(state, 1);
+
+    if (value != Py_None)
     {
         PyObject *message = PyObject_Str(value);
 
         if (message != NULL)
         {
-            jfieldID fid =
-                vm_env->GetFieldID(jcls, "message", "Ljava/lang/String;");
             jstring str = env->fromPyString(message);
-
-            vm_env->SetObjectField(self, fid, str);
-            vm_env->DeleteLocalRef(str);
             Py_DECREF(message);
+
+            return str;
         }
     }
 
+    return NULL;
+}
+
+static jstring JNICALL _PythonException_getErrorTraceback(JNIEnv *vm_env,
+                                                          jobject self)
+{
+    jclass jcls = vm_env->GetObjectClass(self);
+    jfieldID fid = vm_env->GetFieldID(jcls, "py_error_state", "J");
+    PyObject *state = (PyObject *) vm_env->GetLongField(self, fid);
+
+    if (state == NULL)
+        return NULL;
+
+    PythonGIL gil(vm_env);
     PyObject *module = NULL, *cls = NULL, *stringIO = NULL, *result = NULL;
     PyObject *_stderr = PySys_GetObject("stderr");
+
     if (!_stderr)
-        goto err;
+        return NULL;
 
     module = PyImport_ImportModule("cStringIO");
     if (!module)
-        goto err;
+        return NULL;
 
     cls = PyObject_GetAttrString(module, "StringIO");
     Py_DECREF(module);
     if (!cls)
-        goto err;
+        return NULL;
 
     stringIO = PyObject_CallObject(cls, NULL);
     Py_DECREF(cls);
     if (!stringIO)
-        goto err;
+        return NULL;
 
     Py_INCREF(_stderr);
     PySys_SetObject("stderr", stringIO);
 
+    PyObject *type = PyTuple_GET_ITEM(state, 0);
+    PyObject *value = PyTuple_GET_ITEM(state, 1);
+    PyObject *tb = PyTuple_GET_ITEM(state, 2);
+    jstring str = NULL;
+
+    Py_INCREF(type);
+    if (value == Py_None)
+        value = NULL;
+    else
+        Py_INCREF(value);
+    if (tb == Py_None)
+        tb = NULL;
+    else
+        Py_INCREF(tb);
+
     PyErr_Restore(type, value, tb);
     PyErr_Print();
 
@@ -729,39 +924,31 @@ static void JNICALL _PythonException_getErrorInfo(JNIEnv *vm_env, jobject self)
 
     if (result != NULL)
     {
-        jfieldID fid =
-            vm_env->GetFieldID(jcls, "traceback", "Ljava/lang/String;");
-        jstring str = env->fromPyString(result);
-
-        vm_env->SetObjectField(self, fid, str);
-        vm_env->DeleteLocalRef(str);
+        str = env->fromPyString(result);
         Py_DECREF(result);
     }
 
     PySys_SetObject("stderr", _stderr);
     Py_DECREF(_stderr);
 
-    return;
-
-  err:
-    PyErr_Restore(type, value, tb);
-}
-
-static void JNICALL _PythonException_clear(JNIEnv *vm_env, jobject self)
-{
-    PythonGIL gil(vm_env);
-    PyErr_Clear();
+    return str;
 }
 
 static void registerNatives(JNIEnv *vm_env)
 {
     jclass cls = vm_env->FindClass("org/apache/jcc/PythonException");
     JNINativeMethod methods[] = {
-        { "getErrorInfo", "()V", (void *) _PythonException_getErrorInfo },
-        { "clear", "()V", (void *) _PythonException_clear },
+        { "pythonDecRef", "()V", (void *) _PythonException_pythonDecRef },
+        { "saveErrorState", "()V", (void *) _PythonException_saveErrorState },
+        { "getErrorName", "()Ljava/lang/String;",
+          (void *) _PythonException_getErrorName },
+        { "getErrorMessage", "()Ljava/lang/String;",
+          (void *) _PythonException_getErrorMessage },
+        { "getErrorTraceback", "()Ljava/lang/String;",
+          (void *) _PythonException_getErrorTraceback },
     };
 
-    vm_env->RegisterNatives(cls, methods, 2);
+    vm_env->RegisterNatives(cls, methods, sizeof(methods) / sizeof(methods[0]));
 }
 
 #endif /* _jcc_lib */
